#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdarg.h>
#include <windows.h>
#include <richedit.h>
#include <libwcurses.h>

// Warum kein Prototyp????
int vasprintf(char**, const char*, va_list);

extern WINDOW *stdscr;
extern volatile HWND mainWnd;
extern int echo_enable;
extern int last_key;

extern const int LINES;
extern const int COLS;
extern HDC hdc;

int halfdelay_var=-1;


#define GET_CHAR(__X, __Y) cons_text[(__X)+(__Y)*CONS_WIDTH]
#define SET_CHAR(__X, __Y, __C) do { \
        cons_text[(__X)+(__Y)*CONS_WIDTH]=__C; \
    } while(0)

#define CONS_ATTR(__X, __Y) cons_attrs[(__X)+(__Y)*CONS_WIDTH]



int cons_color_pairs[256];
uint32_t cons_attrs[CONS_WIDTH*CONS_HEIGHT];
extern char cons_text[CONS_WIDTH*CONS_HEIGHT+1];


int color_r[]={0, 255, 0, 255, 0, 255, 0, 255};
int color_g[]={0, 0, 255, 255, 0, 0, 255, 255};
int color_b[]={0, 0, 0, 0, 255, 255, 255, 255};

void win_textout(HDC dc, WINDOW *win) {
    int i, j, s;
    TEXTMETRIC tm;
    int cpair, fg, bg, old_fg, old_bg;
    int x0, y0, x1, y1;

    x0=win->_begx;
    y0=win->_begy;
    x1=win->_begx+win->_maxx;
    y1=win->_begy+win->_maxy;

    GetTextMetrics(dc, &tm);

    for (i=y0;i<y1;i++) {
        s=0;

        if (CONS_ATTR(0, i)&CURSES_ATTR_UNSET) {
            old_bg=COLOR_BLACK;
            old_fg=COLOR_WHITE;
        } else {
            cpair=cons_color_pairs[CONS_ATTR(0, i)&0xff];
            old_fg=(cpair>>4)&0x0f;
            old_bg=cpair&0x0f;
        }
        
        SetTextColor(dc, RGB(color_r[old_fg], color_g[old_fg], color_b[old_fg]));
        SetBkColor(dc, RGB(color_r[old_bg], color_g[old_bg], color_b[old_bg]));

        for (j=x0;j<x1;j++) {

            if (CONS_ATTR(j, i)&CURSES_ATTR_UNSET) {
                fg=COLOR_WHITE;
                bg=COLOR_BLACK;
            } else {
                cpair=cons_color_pairs[CONS_ATTR(j, i)&0xff];
                fg=(cpair>>4)&0x0f;
                bg=cpair&0x0f;
            }


            if (fg!=old_fg || bg!=old_bg) {
                TextOut(dc, 10+s*tm.tmAveCharWidth, 10+i*tm.tmHeight, cons_text+i*CONS_WIDTH+s, j-s);
                s=j;
            }

            if (fg!=old_fg) {
                SetTextColor(dc, RGB(color_r[fg], color_g[fg], color_b[fg]));
                old_fg=fg;
            }
            if (bg!=old_bg) {
                SetBkColor(dc, RGB(color_r[bg], color_g[bg], color_b[bg]));
                old_bg=bg;
            }
        }
        if ((j-s)>0) {
            TextOut(dc, 10+s*tm.tmAveCharWidth, 10+i*tm.tmHeight,cons_text+i*CONS_WIDTH+s, j-s);
        }
    }
}


static inline void cons_update(WINDOW *win) {
    HDC dc;

    dc=GetDC(mainWnd);

    SelectObject(dc, GetStockObject(ANSI_FIXED_FONT));
    win_textout(dc, win);

    ReleaseDC(mainWnd, dc);
}

static inline int str2window(WINDOW *win, const char *str) {
    int i, j, x, y;

    x=win->_begx+win->_maxx;
    y=win->_begy+win->_maxy;

    i=win->_begx+win->_curx;
    j=win->_begy+win->_cury;
    

    while (*str!='\0') {

        switch (*str) {

            case '\r':
                break;

            case '\n':
                j++;
                str++;
                i=win->_begx;
                win->_curx=win->_begx;
                if (j>y) {
                    return (*str!='\0'?__LINE__:0);
                }
                win->_cury++;
                break;

            default:
                SET_CHAR(i, j, *str);
                if (win->_attrs==CURSES_ATTR_UNSET) {
                    CONS_ATTR(i, j)=-1;
                } else {
                    CONS_ATTR(i, j)=win->_attrs;
                }
                str++;
                i++;
                win->_curx++;
                if (i>x) {
                    i=win->_begx;
                    win->_curx=win->_begx;
                    j++;
                    if (j>y) {
                        return (*str!='\0'?__LINE__:0);
                    }
                    win->_cury++;
                }
                break;
        }
    }
    
    return 0;
}

int getch(void) {
    return wgetch(stdscr);
}


int wgetch(WINDOW *win) {
    int r;
    DWORD tc0, tc1, tcd;

    if (win->_delay_ms>=0) {
        int dms=win->_delay_ms;

        tc0=GetTickCount();
        while (dms>0) {
            Sleep(1);
            tc1=GetTickCount();
            if (tc1>tc0) {
                tcd=tc1-tc0;
            } else {
                tcd=0xffffffff-tc0+tc1;
            }
            dms=dms-tcd;
        }
        if (last_key==-1) {
            return ERR;
        }
        r=last_key;
        last_key=-1;

        return r;
    }

    if (halfdelay_var!=-1) {
        tc0=GetTickCount();

        do {
            Sleep(50);
            tc1=GetTickCount();
            if (tc1>tc0) {
                tcd=tc1-tc0;
            } else {
                tcd=0xffffffff-tc0+tc1;
            }
            if (last_key!=-1) {
                break;
            }
        } while (tcd<(100*halfdelay_var));

        if (last_key==-1) {
            return ERR;
        }
        
        r=last_key;
        last_key=-1;

        return r;
    }

    if (win->_nodelay==TRUE) {
        if (last_key==-1) {
            return ERR;
        }
        r=last_key;
        last_key=-1;

        return r;
    }

    last_key=-1;
    while (last_key==-1) {
        Sleep(50);
    }
    r=last_key;
    last_key=-1;

    return r;
}

int beep(void) {
    Beep(1000, 250);

    return OK;
}

int echo(void) {
    echo_enable=1;
    return OK;
}

int noecho(void) {
    echo_enable=0;

    return OK;
}

WINDOW *initscr(void) {
    int i;
    for (i=0;i<(sizeof(cons_attrs)/sizeof(int));i++) {
        cons_attrs[i]=-1;
    }
    return stdscr;
}

int refresh(void) {
    return wrefresh(stdscr);
}

int wrefresh(WINDOW *win) {
    cons_update(stdscr);

    return OK;
}

int endwin(void) {
    return OK;
}

int mvwprintw(WINDOW *win, int y, int x, const char *fmt, ...) {
    va_list ap;

    if (x<0||x>win->_maxx||y<0||y>win->_maxy) {
        return ERR;
    }

    va_start(ap, fmt);
    win->_curx=x;
    win->_cury=y;
    vw_printw(win, fmt, ap);
    va_end(ap);

    return OK;
}

int mvprintw(int y, int x, const char *fmt, ...) {
    va_list a;

    va_start(a, fmt);

    stdscr->_curx=x;
    stdscr->_cury=y;
    vw_printw(stdscr, fmt, a);
    
    va_end(a);

    return OK;
}

int vw_printw(WINDOW *win, const char *fmt, va_list ap) {
    char *str;

    vasprintf(&str, fmt, ap);
    str2window(win, str);
    
    free(str);

    return OK;
}

int wprintw(WINDOW *win, const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vw_printw(win, fmt, ap);
    va_end(ap);

    return OK;
}

int printw(const char *fmt, ...) {
    va_list ap;

    va_start(ap, fmt);
    vw_printw(stdscr, fmt, ap);
    va_end(ap);

    return OK;
}

WINDOW* newwin(int nlines, int ncols, int begin_y, int begin_x) {
    WINDOW *ret;
    
    ret=malloc(sizeof(WINDOW));

    ret->_begx=begin_x;
    ret->_begy=begin_y;
    ret->_maxx=ncols?(ncols-1):(COLS-begin_x-1);
    ret->_maxy=nlines?(nlines-1):(LINES-begin_y-1);
    ret->_curx=0;
    ret->_cury=0;
    ret->_attrs=CURSES_ATTR_UNSET;
    ret->_bkgrnd=' ';
    ret->_nodelay=FALSE;
    ret->_delay_ms=-1;

    werase(ret);

    return ret;
}

int mvwin(WINDOW *win, int y, int x) {
    win->_begx=x;
    win->_begy=y;

    return OK;
}

int wresize(WINDOW *win, int lines, int columns) {
    win->_maxx=columns-1;
    win->_maxy=lines-1;
    return OK;
}

int flash(void) {
    return beep();
}

int delwin(WINDOW *win) {
    free(win);

    return OK;
}

int werase(WINDOW *win) {
    int i, j, x, y;

    if (win==NULL) {
        return ERR;
    }

    x=win->_begx+win->_maxx;
    y=win->_begy+win->_maxy;

    for (i=win->_begx;i<=x;i++) {
        for (j=win->_begy;j<=y;j++) {
            SET_CHAR(i, j, ' ');
            CONS_ATTR(i, j)=-1;
        }
    }
    win->_curx=0;
    win->_cury=0;

    return OK;
}

int erase(void) {

    return clear();
}

int box(WINDOW *win, char verch, char horch) {
    int i;
    int w, h;    

    w=win->_maxx;
    h=win->_maxy;

    for (i=0;i<=w;i++) {
        SET_CHAR(win->_begx+i, win->_begy, '-');
        SET_CHAR(win->_begx+i, win->_begy+h, '-');
    }

    for (i=1;i<=h;i++) {
        SET_CHAR(win->_begx, win->_begy+i, '|');
        SET_CHAR(win->_begx+w,win->_begy+i, '|');
    }

    SET_CHAR(win->_begx, win->_begy, '+');
    SET_CHAR(win->_begx+w, win->_begy, '+');
    SET_CHAR(win->_begx, win->_begy+h, '+');
    SET_CHAR(win->_begx+w, win->_begy+h, '+');

    return OK;
}

int wmove(WINDOW *win, int y, int x) {
    win->_curx=x;
    win->_cury=y;

    return OK;
}

int move(int y, int x) {
    return wmove(stdscr, y, x);
}

int mvwaddch(WINDOW *win, int y, int x, char ch) {
    wmove(win, y, x);

    SET_CHAR(win->_begx+win->_curx, win->_begy+win->_cury, ch);
    CONS_ATTR(win->_begx+win->_curx, win->_begy+win->_cury)=win->_attrs;

    if (win->_curx>win->_maxx) {
        if (win->_cury>win->_maxy) {
            return OK;
        }
        win->_cury++;
        win->_curx=0;
        return OK;
    }
    win->_curx++;

    return OK;
}

int mvaddch(int y, int x, char ch) {
    return mvwaddch(stdscr, y, x, ch);
}

int waddch(WINDOW *win, const char ch) {
    return mvwaddch(win, win->_cury, win->_curx, ch);
}

int start_color(void) {
    return OK;
}

int has_colors(void) {
    return TRUE;
}

int curs_set(int visibility) {
    return 0;
}


int init_pair(short color, short f, short b) {


    if (f==-1) {
        f=COLOR_WHITE;
    }
    if (b==-1) {
        b=COLOR_BLACK;
    }
    
    cons_color_pairs[color&0xff]=((f&0x0f)<<4)|(b&0x0f);
    return OK;
}

int wbkgd(WINDOW *win, char ch) {
    return OK;
}

int bkgd(char ch) {
    return wbkgd(stdscr, ch);
}

int use_default_colors(void) {
    return OK;
}

int keypad(WINDOW *win, bool bf) {
    return OK;
}

int wattrset(WINDOW *win, int attrs) {
    win->_attrs=attrs;
    return OK;
}

int wclear(WINDOW *win) {

    werase(win);

    return OK;
}

int clear(void) {
    werase(stdscr);
    return OK;
}

int nodelay(WINDOW *win, bool bf) {
    win->_nodelay=bf;

    return OK;
}

chtype mvwinch(WINDOW *win, int y, int x) {
    return GET_CHAR(win->_begx+x, win->_begy+y);
}

chtype winch(WINDOW *win) {
    return mvwinch(win, win->_cury, win->_curx);
}

chtype mvinch(int y, int x) {
    return mvwinch(stdscr, y, x);
}

chtype inch(void) {
    return mvwinch(stdscr, stdscr->_cury, stdscr->_curx);
}

int cbreak(void) {
    return OK;
}

int intrflush(WINDOW *win, bool bf) {
    return OK;
}

int nonl(void) {
    return OK;
}

int wcolor_set(WINDOW *win, short pair, void* opts) {
    return OK;
}

int color_set(short pair, void* opts) {
    return OK;
}


int wattron(WINDOW *win, int attrs) {
    win->_attrs=(attrs)&(~CURSES_ATTR_UNSET);
    return OK;
}

int attron(int attrs) {
    return wattron(stdscr, attrs);
}

int wattroff(WINDOW *win, int attrs) {
    win->_attrs=win->_attrs^attrs;
    return OK;
}

int attroff(int attrs) {
    return wattron(stdscr, attrs);
}

void wtimeout(WINDOW *win, int delay) {
    win->_delay_ms=delay; 
}

int notimeout(WINDOW *win, bool bf) {
    return OK;
}

#define STR_CHUNK_SZ 79 // 80 Byte in 'data' + 8 Byte 'next' = 64bit aligned :)
struct str_chunk {
    char data[(STR_CHUNK_SZ)+1];
    struct str_chunk *next;
};
int mvwscanw(WINDOW *win, int y, int x, const char *fmt, ...) {
    int len, c;
    struct str_chunk *chk, *t;
    char *str;
    int ret;
    va_list va;

    va_start(va, fmt);

    wmove(win, y, x);
    c=0;
    chk=malloc(sizeof(struct str_chunk));
    chk->data[0]='\0';
    chk->next=NULL;

    last_key=-1;

    //TODO: echo/noecho? getch verwenden?
    while (1) {

        Sleep(50);
        if (last_key!=-1) {
            if (last_key=='\r'||last_key=='\n') {
                break;
            } else if (last_key==KEY_BACKSPACE) {
                last_key=-1; //TODO: Backspace: Zeichen loeschen + Byte vom Bildschirm nehmen.
                continue;
            } else {
                printw("%c", last_key); // TODO: nur, wenn echo on ist.
            }

            len=strlen(chk->data);
            chk->data[len]=last_key;
            chk->data[len+1]='\0';

            if ((len+1) == STR_CHUNK_SZ) {
                t=malloc(sizeof(struct str_chunk));
                t->next=chk;
                t->data[0]='\0';
                chk=t;
                c++;
            }

            last_key=-1;
        }
    }

    str=malloc((c+1)*STR_CHUNK_SZ);
    str[strlen(chk->data) + c*STR_CHUNK_SZ]='\0';

    do {
        memcpy(str+c*STR_CHUNK_SZ, chk->data, STR_CHUNK_SZ);
        t=chk->next;
        free(chk);
        chk=t;
    } while (c--);

    ret=vsscanf(str, fmt, va);
    printw("\n");

    free(str);
    va_end(va);

    return ret;
}

int whline(WINDOW *win, chtype ch, int n) {
    int i;

    for (i=0;i<win->_maxx && i<n;i++) {
        SET_CHAR(win->_begx+i, win->_begy, ch);
    }
    return OK;
}

int wborder(WINDOW *win, chtype ls, chtype rs, chtype ts, chtype bs, chtype	tl, chtype tr, chtype bl, chtype br) {
    box(win, 0, 0);

    return OK;
}

int border(chtype ls, chtype rs, chtype ts, chtype bs, chtype tl, chtype tr, chtype bl, chtype br) {
    return wborder(stdscr, ls, rs, ts, bs, tl, tr, bl, br);
}

int mvwaddstr(WINDOW *win, int y, int x, const char *str) {
    if (win==NULL || str==NULL) {
        return ERR;
    }

    win->_curx=x;
    win->_cury=y;
    str2window(win, str);

    return OK;
}

int waddstr(WINDOW *win, const char *str) {
    return mvwaddstr(stdscr, stdscr->_begx, stdscr->_begy, str);
}

int wnoutrefresh(WINDOW *win) {
    return OK;
}

int doupdate(void) {
    cons_update(stdscr);
    return OK;
}

int scrollok(WINDOW *win, bool bf) {
    return OK;
}

WINDOW *derwin(WINDOW *orig, int nlines, int ncols, int begin_y, int begin_x) {
    return newwin(nlines, ncols, orig->_begy+begin_y, orig->_begx+begin_x);
}

int wstandend(WINDOW *win) {
    return OK;
}

int wstandout(WINDOW *win) {
    return OK;
}

int pair_content(short pair, short *f, short *b) {
    if (pair&0xff00) {
        return ERR;
    }
    *b=cons_color_pairs[pair]&0x0f;
    *f=(cons_color_pairs[pair]>>4)&0x0f;

    return OK;
}

void wbkgdset(WINDOW *win, chtype ch) {
    win->_bkgrnd=ch;
}

int touchwin(WINDOW *win) {
    return OK;
}

int flushinp(void) {
    last_key=-1;
    return OK;
}

int halfdelay(int tenths) {
    halfdelay_var=tenths;
    return OK;
}

int wvline(WINDOW *win, chtype ch, int n) {
    int i, y_max;

    y_max=win->_maxy<n?win->_maxy:n;
    for (i=0;i<y_max;i++) {
        SET_CHAR(win->_begx, win->_begy+i, ch);
    }

    return OK;
}
