#ifndef __LIBWCURSES_H__
#define __LIBWCURSES_H__

#include <stdint.h>
#include <stdarg.h>
#include <stdbool.h>

#define CONS_WIDTH 81
#define CONS_HEIGHT 26

#undef TRUE
#define TRUE 1

#undef FALSE
#define FALSE 0

#undef OK
#define OK 0

#undef ERR
#define ERR (-1)

#define COLORS 8
#define COLOR_PAIR(__N) (__N)

#define A_BOLD 0//(1<<21)
#define A_REVERSE 0//(1<<18)

#define ACS_HLINE '-'
#define ACS_VLINE '|'


#define CURSES_ATTR_UNSET 0x80000000

// Farben
#define COLOR_BLACK     0
#define COLOR_RED       1
#define COLOR_GREEN     2
#define COLOR_YELLOW    3
#define COLOR_BLUE      4
#define COLOR_MAGENTA   5
#define COLOR_CYAN      6
#define COLOR_WHITE     7
#define COLOR_PAIRS     64

#define KEY_DOWN	0402
#define KEY_UP		0403
#define KEY_LEFT	0404
#define KEY_RIGHT	0405
#define KEY_BACKSPACE   0407


#define KEY_ENTER 0x157

#define KEY_PPAGE 0x153
#define KEY_NPAGE 0x152
#define KEY_END 0x168
#define KEY_HOME 0x106
#define KEY_A1 0x15c
#define KEY_A3 0x15d
#define KEY_C1 0x15f
#define KEY_C3 0x160
#define KEY_B2 0x15e
#define KEY_IC 0x14b
#define KEY_REFRESH 0x175
#define KEY_HELP 0x16b
#define KEY_BREAK 0x101
#define KEY_SUSPEND 0x197

#define ACS_ULCORNER 0x40006c
#define ACS_URCORNER 0x40006b
#define ACS_LLCORNER 0x40006d
#define ACS_LRCORNER 0x40006a
#define ACS_UARROW 0x40002d
#define ACS_DARROW 0x40002e
#define ACS_RARROW 0x40002b
#define ACS_LARROW 0x40002c
#define ACS_DIAMOND 0x400060


typedef uint32_t chtype;
typedef int uid_t;

#define getmaxyx(__W, __Y, __X) do { \
        (__X)=(__W)->_maxx; \
        (__Y)=(__W)->_maxy; \
    } while(0)

typedef struct {
    size_t _curx, _cury;
    size_t _maxx, _maxy;
    size_t _begx, _begy;
    uint32_t _attrs;
    int _nodelay;
    int _delay_ms;
    chtype _bkgrnd;
} WINDOW;

int attroff(int attrs);
int attron(int attrs);
int beep(void);
int bkgd(char ch);
int border(chtype ls, chtype rs, chtype ts, chtype bs, chtype tl, chtype tr, chtype bl, chtype br);
int box(WINDOW *win, char verch, char horch);
int cbreak(void);
int clear(void);
int color_set(short pair, void* opts);
int curs_set(int visibility);
int delwin(WINDOW *win);
int doupdate(void);
WINDOW *derwin(WINDOW *orig, int nlines, int ncols, int begin_y, int begin_x);
int echo(void);
int endwin(void);
int erase(void);
int flash(void);
int flushinp(void);
int getch(void);
int halfdelay(int tenths);
int has_colors(void);
chtype inch(void);
WINDOW *initscr(void);
int init_pair(short color, short f, short b);
int intrflush(WINDOW *win, bool bf);
int keypad(WINDOW *win, bool bf);
int move(int y, int x);
int mvaddch(int y, int x, char ch);
chtype mvinch(int y, int x);
int mvprintw(int y, int x, const char *fmt, ...);
int mvwaddch(WINDOW *win, int y, int x, char ch);
int mvwaddstr(WINDOW *win, int y, int x, const char *str);
int mvwin(WINDOW *win, int y, int x);
chtype mvwinch(WINDOW *win, int y, int x);
int mvwprintw(WINDOW *win, int y, int x, const char *fmt, ...);
int mvwscanw(WINDOW *win, int y, int x, const char *fmt, ...);
WINDOW* newwin(int nline, int ncols, int begin_y, int begin_x);
int nodelay(WINDOW *win, bool bf);
int noecho(void);
int nonl(void);
int notimeout(WINDOW *win, bool bf);
int pair_content(short pair, short *f, short *b);
int printw(const char *ftm, ...);
int refresh(void);
int scrollok(WINDOW *win, bool bf);
int start_color(void);
int touchwin(WINDOW *win);
int use_default_colors(void);
int vw_printw(WINDOW *win, const char *fmt, va_list ap);
int waddch(WINDOW *win, const char ch);
int waddstr(WINDOW *win, const char *str);
int wattroff(WINDOW *win, int attrs);
int wattron(WINDOW *win, int attrs);
int wattrset(WINDOW *win, int attrs);
int wbkgd(WINDOW *win, char ch);
void wbkgdset(WINDOW *win, chtype ch);
int wborder(WINDOW *win, chtype ls, chtype rs, chtype ts, chtype bs, chtype	tl, chtype tr, chtype bl, chtype br);
int wclear(WINDOW *win);
int wcolor_set(WINDOW *win, short pair, void* opts);
int werase(WINDOW *win);
int wgetch(WINDOW *win);
int whline(WINDOW *win, chtype ch, int n);
chtype winch(WINDOW *win);
int wmove(WINDOW *win, int y, int x);
int wnoutrefresh(WINDOW *win);
int wprintw(WINDOW *win, const char *fmt, ...);
int wrefresh(WINDOW *win);
int wresize(WINDOW *win, int lines, int columns);
int wstandend(WINDOW *win);
int wstandout(WINDOW *win);
void wtimeout(WINDOW *win, int delay);
int wvline(WINDOW *win, chtype ch, int n);

#endif

