#include <stdio.h>
#include <windows.h>
#include <string.h>
#include <richedit.h>
#include <stdarg.h>
#include <libwcurses.h>


DWORD WINAPI updateRtfWin(LPVOID x);
void win_textout(HDC dc, WINDOW *win);

int _main(int, char**);

struct cmdArgs {
    int argc;
    char **argv;
};

struct cmdArgs cmdArgs;
DWORD threadId;
LPSTR args;
char exeFile[PATH_MAX];
volatile HWND mainWnd;
int last_key=-1;

char cons_text[CONS_WIDTH*CONS_HEIGHT+1];

volatile WINDOW stdscr0;
WINDOW *stdscr;
const int LINES = CONS_HEIGHT-1;
const int COLS = CONS_WIDTH-1;

int echo_enable=0;



static inline void cons_init() { // newwin()?
    memset(cons_text, ' ', CONS_WIDTH*CONS_HEIGHT);
    stdscr=(WINDOW*)&stdscr0;
    stdscr->_begx=0;
    stdscr->_begy=0;
    stdscr->_curx=0;
    stdscr->_cury=0;
    stdscr->_maxx=CONS_WIDTH-1;
    stdscr->_maxy=CONS_HEIGHT-1;
    stdscr->_attrs=CURSES_ATTR_UNSET;
    stdscr->_bkgrnd=' ';
    stdscr->_nodelay=0;
    stdscr->_delay_ms=-1;
}


DWORD WINAPI threadProc(LPVOID p0) {
    int ret;
    char **argv;
    char *argstr;
    int i, len, quot, words;
    
    len=strlen(args);
    argstr=malloc(len+1);
    strcpy(argstr, args);
    quot=0;
    words=1;
    if (len) {
        words=2;
    }
    for (i=0;i<len;i++) {
        if (quot) {
            if (argstr[i]=='"') {
                quot=0;
            }
        } else {
            if (argstr[i]=='"') {
                quot=1;
            }
            if (argstr[i]==' ') {
                words++;
            }
        }
    }

    argv=malloc(words*sizeof(char*));
    argv[0]=exeFile;
    words=1;
    quot=0;

    if (len) {
        words=2;
        argv[1]=argstr;

        for (i=0;i<len;i++) {
            if (quot) {
                if (argstr[i]=='"') {
                    quot=0;
                }
            } else {
                if (argstr[i]=='"') {
                    quot=1;
                }
                if (argstr[i]==' ') {
                    argstr[i]='\0';
                    argv[words++]=argstr+i+1;
                }
            }
        }
    }

    ret=_main(words, argv);
    free(argv);
    
    clear();
    printw("\n\n> main() returned %i", ret);
    
    return 0;
}

LRESULT CALLBACK wndProc(HWND wnd, UINT msg, WPARAM wp, LPARAM lp) {

    switch (msg) {
        
        case WM_CREATE: {
            cons_init();
            CreateThread(NULL, 0, threadProc, NULL, 0, &threadId);
            SetFocus(mainWnd);
    
            break;
        }

        case WM_PAINT: {
            RECT r;
            PAINTSTRUCT ps;
            HBRUSH hbr;

            GetClientRect(wnd, &r);
            BeginPaint(wnd, &ps);
            hbr=CreateSolidBrush(RGB(0, 0, 0));
            SelectObject(ps.hdc, hbr);
            Rectangle(ps.hdc, 0, 0, r.right, r.bottom);
            SelectObject(ps.hdc, GetStockObject(ANSI_FIXED_FONT));
            win_textout(ps.hdc, stdscr);
            DeleteObject(hbr);
            EndPaint(wnd, &ps);
            break;
        }

        case WM_KEYDOWN: {
            switch (wp) {
                case VK_UP:
                    last_key=KEY_UP;
                    break;
                case VK_DOWN:
                    last_key=KEY_DOWN;
                    break;
                case VK_LEFT:
                    last_key=KEY_LEFT;
                    break;
                case VK_RIGHT:
                    last_key=KEY_RIGHT;
                    break;

                default:
                    break;
            }
            break;
        }

        case WM_CHAR: {
            
            char c=wp;
            if (c==8) { // ASCII-Backspace.
                last_key=KEY_BACKSPACE;
                break;
            }
            if (c=='\r') { // CR-LF??
                c='\n';
                last_key=c;
                break;

            }
            last_key=c;

            break;
        }

        case WM_CLOSE: {
            PostQuitMessage(0);
            break;
        }

        case WM_DESTROY: {
            SendMessage(wnd, WM_CLOSE, 0, 0);
            break;
        }

        case WM_KILLFOCUS: {
            SetFocus(mainWnd);
            break;
        }
        default:
            return DefWindowProc(wnd, msg, wp, lp);
    }
    return 0;
}

int WINAPI WinMain(HINSTANCE hThisInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd) {
    WNDCLASS wc;
    MSG msg;

    char *p;
    char winttl[PATH_MAX+20];
    char *wndClass="main.wincc";


    GetModuleFileName(NULL, exeFile, MAX_PATH);
    p=exeFile+strlen(exeFile);
    args=lpCmdLine;


    while (*p!='\\' && p!=exeFile) {
        p--;
    }
    p=*p=='\\'?p+1:p; // Falls exeFile=="name.exe"

    wc.cbClsExtra=0;
    wc.cbWndExtra=0;
    wc.hbrBackground=CreateSolidBrush(RGB(0, 0, 0));
    wc.hCursor=LoadCursor(0, IDC_ARROW);
    wc.hIcon=LoadIcon(0, IDI_APPLICATION);
    wc.hInstance=hThisInstance;
    wc.lpfnWndProc=wndProc;
    wc.lpszClassName=wndClass;
    wc.lpszMenuName=NULL;
    wc.style=0;

    RegisterClass(&wc);
    sprintf(winttl, "WinCC 1.0 - %s", p);
    mainWnd=CreateWindow(
        wndClass,
        winttl,
        WS_OVERLAPPEDWINDOW,
        CW_USEDEFAULT,
        CW_USEDEFAULT,
        700,
        500,
        HWND_DESKTOP,
        NULL,
        hThisInstance,
        0
    );

    ShowWindow(mainWnd, nShowCmd);
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    return 0;
}
